"""Collectors specific to OS X"""

from rekall.plugins.collectors.darwin import allocations
from rekall.plugins.collectors.darwin import common
from rekall.plugins.collectors.darwin import handles
from rekall.plugins.collectors.darwin import networking
from rekall.plugins.collectors.darwin import processes
from rekall.plugins.collectors.darwin import sessions
from rekall.plugins.collectors.darwin import zones
