"""Collectors specific to Windows"""


from rekall.plugins.collectors.windows import common
from rekall.plugins.collectors.windows import handles
from rekall.plugins.collectors.windows import networking
from rekall.plugins.collectors.windows import pool
from rekall.plugins.collectors.windows import processes
